/*
 * Copyright (C) 2014 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: 14 April, 2014
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */



#ifndef CFLUXSOLVER_FULLSWOF_HPP_
#define CFLUXSOLVER_FULLSWOF_HPP_


template <typename T>
class CFluxSolver_FullSWOF
{
public:
	/**
	 * \brief compute the flux and store the result to o_flux
	 *
	 * this method uses a upwinding method based on the velocity component in combination with a numerical
	 * diffusion similar to the lax-friedrichs flux
	 *
	 * the input data is assumed to be rotated to the edge normal pointing along the positive part of the x axis
	 */
	inline void op_edge_edge(
			const CSimulationNodeData &i_edgeData_left,		///< edge data on left (left) edge
			const CSimulationNodeData &i_edgeData_right,		///< edge data on right (outer) edge

			CSimulationNodeData *o_edgeFlux_left,		///< output for left flux
			CSimulationNodeData *o_edgeFlux_right,		///< output for outer flux

			T *o_max_wave_speed_left,				///< maximum wave speed
			T *o_max_wave_speed_right,				///< maximum wave speed
			T i_gravitational_constant				///< gravitational constant
	)
	{
		T hl_0 = i_edgeData_left.h;
		T hr_0 = i_edgeData_right.h;
		T dz = i_edgeData_right.b - i_edgeData_left.b;

		/**
		 * the following codes originates from FULL SWOF
		 */

		// from FULLSWOF, hydrostatic.cpp
		// hydrostatic height
		T h_L = std::max((T)0, hl_0-std::max((T)0,dz));
		T h_R = std::max((T)0, hr_0-std::max((T)0,-dz));



		// from FULLSWOF, f_hll2.cpp
		/**
		 * @details
		 * If the water heights on the two sides are small, there is no water.
		 * Else, HLL reduced formulation is used (see \cite Bouchut04):
		 * \f[{\cal F}(U_L,U_R)=t_{1}F(U_R)+t_{2}F(U_L)-t_{3}(U_R-U_L),\f]
		 * with \f[t_1=\frac{\min(c_2,0)-\min(c_1,0)}{c_2-c_1},\quad t_2=1-t_1,\quad t_3=\frac{c_2|c_1|-c_1|c_2|}{2(c_2-c_1)},\f]
		 * \f[c_{1}={\inf\limits_{U=U_L,U_R}}\left({\inf\limits_{j\in\{1,2\}}}|\lambda_{j}(U)|\right) \mbox{ and }
		 * c_{2}={\sup\limits_{U=U_L,U_R}}\left({\sup\limits_{j\in\{1,2\}}}|\lambda_{j}(U)|\right),\f]
		 * where \f$\lambda_1(U)=u-\sqrt{gh}\f$ and \f$\lambda_2(U)=u+\sqrt{gh}\f$ are the eigenvalues of the Shallow Water system,
		 * \f$U = {}^t (h,hu)\f$ and \f$ F(U) = {}^t (hu, hu^2+gh^2/2) \f$.
		 * @param[in] h_L water height at the left of the interface where the flux is calculated.
		 * @param[in] u_L velocity at the left of the interface where the flux is calculated.
		 * @param[in] h_R water height at the right of the interface where the flux is calculated.
		 * @param[in] u_R velocity at the right of the interface where the flux is calculated.
		 * @par Modifies
		 * Flux#f1 first componant of the numerical flux.\n
		 * Flux#f2 second componant of the numerical flux.\n
		 * Flux#cfl value of the CFL.
		 */

		const T HE_CA = 1.e-10;
		const T EPSILON = 1.e-10;
		const T GRAV = 9.81;

		T f1, f2, cfl;
		T u_L = i_edgeData_left.hu;
		T u_R = i_edgeData_right.hu;
		T tx = 1; // value of dt/dx
		typedef T SCALAR;

		if (h_L<HE_CA && h_R<HE_CA)
		{
			f1 = 0.;
			f2 = 0.;
			cfl = 0.;
		}
		else
		{
#if 1

			// HLL
			SCALAR grav_h_L = GRAV*h_L;
			SCALAR grav_h_R = GRAV*h_R;
			SCALAR q_R = u_R*h_R;
			SCALAR q_L = u_L*h_L;
			SCALAR c1 = std::min(u_L-std::sqrt(grav_h_L),u_R-std::sqrt(grav_h_R));		// as u-sqrt(grav_h) <= u+sqrt(grav_h)
			SCALAR c2 = std::max(u_L+std::sqrt(grav_h_L),u_R+std::sqrt(grav_h_R));		// as u+sqrt(grav_h) >= u-sqrt(grav_h)

			if (std::abs(c1)<EPSILON && std::abs(c2)<EPSILON){		//dry state
				f1 = 0.;
				f2 = 0.;
				cfl = 0.;
			}else if(c1>=EPSILON){			//supercritical flow from left to right
				f1 = q_L;
				f2 = q_L*u_L+(grav_h_L*h_L)*0.5;
				cfl = c2*tx;				//MAX(fabs(c2),fabs(c1))=c2>0
			}else if(c2<=-EPSILON){			//supercritical flow from right to left
				f1 = q_R;
				f2 = q_R*u_R+(grav_h_R*h_R)*0.5;
				cfl = -c1*tx;				//MAX(fabs(c2),fabs(c1))=-c1>0
			}else{							//subcritical flow
				SCALAR tmp = 1./(c2-c1);
				f1 = (c2*q_L-c1*q_R)*tmp+c1*c2*(h_R-h_L)*tmp;
				// long double are used locally to avoid numerical approximations
				f2 = (SCALAR)((long double)c2*((long double)q_L*(long double)u_L+(long double)grav_h_L*(long double)h_L*0.5)-(long double)c1*((long double)q_R*(long double)u_R+(long double)grav_h_R*(long double)h_R*0.5))*(long double)tmp+(long double)c1*(long double)c2*((long double)q_R-(long double)q_L)*(long double)tmp;
				cfl = std::max(std::abs(c1),std::abs(c2))*tx;
			}

#else

			// HLL2
			SCALAR grav_h_L = GRAV*h_L;
			SCALAR grav_h_R = GRAV*h_R;
			SCALAR q_R = u_R*h_R;
			SCALAR q_L = u_L*h_L;
			SCALAR c1 = std::min(u_L-std::sqrt(grav_h_L),u_R-std::sqrt(grav_h_R));		// as u-sqrt(grav_h) <= u+sqrt(grav_h)
			SCALAR c2 = std::max(u_L+std::sqrt(grav_h_L),u_R+std::sqrt(grav_h_R));		// as u+sqrt(grav_h) >= u-sqrt(grav_h)
			SCALAR tmp = 1./(c2-c1);
			SCALAR t1 = (std::min(c2,(T)0)-std::min(c1,(T)0))*tmp;
			SCALAR t2 = 1.-t1;
			SCALAR t3 = (c2*std::abs(c1)-c1*std::abs(c2))*0.5*tmp;

			f1 = t1*q_R+t2*q_L-t3*(h_R-h_L);
			f2 = t1*(q_R*u_R+grav_h_R*h_R*0.5)+t2*(q_L*u_L+grav_h_L*h_L*0.5)-t3*(q_R-q_L);
			cfl = std::max(std::abs(c1),std::abs(c2))*tx;

#endif
		}

		o_edgeFlux_left->h = f1;
		o_edgeFlux_left->hu = f2;
		o_edgeFlux_left->hv = 0;
		o_edgeFlux_left->b = 0;

		o_edgeFlux_right->h = -f1;
		o_edgeFlux_right->hu = -f2;
		o_edgeFlux_right->hv = 0;
		o_edgeFlux_right->b = 0;


		/**
		 * CFL condition
		 */
		*o_max_wave_speed_left = cfl;
		*o_max_wave_speed_right = cfl;
	}


	template <int N>
	void op_edge_edge(
			const CSimulationNodeDataSOA<N> &i_edgeData_left,		///< edge data on left (left) edge
			const CSimulationNodeDataSOA<N> &i_edgeData_right,		///< edge data on right (outer) edge

			CSimulationNodeDataSOA<N> *o_edgeFlux_left,		///< output for left flux
			CSimulationNodeDataSOA<N> *o_edgeFlux_right,		///< output for outer flux

			T *o_max_wave_speed_left,				///< maximum wave speed
			T *o_max_wave_speed_right,				///< maximum wave speed

			T i_gravitational_constant				///< gravitational constant
	)
	{
		T wave_speed_left = 0;
		T wave_speed_right = 0;

		for (int i = 0; i < N; i++)
		{
			T l_wave_speed_left = 0;
			T l_wave_speed_right = 0;

			/*
			 * TODO: SIMD
			 */

			CSimulationNodeData i_left, i_right;

			i_left.h = i_edgeData_left.h[i];
			i_left.hu = i_edgeData_left.hu[i];
			i_left.hv = i_edgeData_left.hv[i];
			i_left.b = i_edgeData_left.b[i];

			i_right.h = i_edgeData_right.h[N-1-i];
			i_right.hu = -i_edgeData_right.hu[N-1-i];
			i_right.hv = -i_edgeData_right.hv[N-1-i];
			i_right.b = i_edgeData_right.b[N-1-i];

			CSimulationNodeData o_left, o_right;

			op_edge_edge(
					i_left,			///< edge data on left (left) edge
					i_right,		///< edge data on right (outer) edge

					&o_left,		///< output for left flux
					&o_right,		///< output for outer flux

					&l_wave_speed_left,				///< maximum wave speed
					&l_wave_speed_right,				///< maximum wave speed

					i_gravitational_constant				///< gravitational constant
			);

			o_edgeFlux_left->h[i] = o_left.h;
			o_edgeFlux_left->hu[i] = o_left.hu;
			o_edgeFlux_left->hv[i] = o_left.hv;
			o_edgeFlux_left->b[i] = o_left.b;

			o_edgeFlux_right->h[N-1-i] = o_right.h;
			o_edgeFlux_right->hu[N-1-i] = -o_right.hu;
			o_edgeFlux_right->hv[N-1-i] = -o_right.hv;
			o_edgeFlux_right->b[N-1-i] = o_right.b;

			wave_speed_left = std::max(wave_speed_left, l_wave_speed_left);
			wave_speed_right = std::max(wave_speed_right, l_wave_speed_right);
		}

		*o_max_wave_speed_left = wave_speed_left;
		*o_max_wave_speed_right = wave_speed_right;
	}
};



#endif /* CFLUXLAXFRIEDRICH_HPP_ */
